"""Comprises tests for the ExplodeColorLayerGlyphsFilter filter."""

from pathlib import Path

import pytest

from ufo2ft.filters.explodeColorLayerGlyphs import ExplodeColorLayerGlyphsFilter
from ufo2ft.util import _GlyphSet


@pytest.fixture
def data_dir():
    return Path(__file__).parent.parent / "data"


def test_strip_color_codepoints(FontClass, data_dir):
    """Test that the filter strips codepoints from glyphs when copying them from
    color layers into default layer alternates.

    See: https://github.com/googlefonts/ufo2ft/pull/739#issuecomment-1516075892"""

    # Load a test UFO with color layers, and give a codepoint to one of the
    # glyphs in those layers.
    ufo = FontClass(data_dir / "ColorTest.ufo")

    color_glyph = ufo.layers["color1"]["a"]
    color_glyph.unicode = 0x3020

    # Apply the filter to the UFO.
    filter = ExplodeColorLayerGlyphsFilter()
    glyphset = _GlyphSet.from_layer(ufo)
    _ = filter(ufo, glyphset)

    # Test that the newly-copied alternate has had its codepoint removed.
    new_default_alt = glyphset["a.color1"]
    assert new_default_alt.unicode is None



from dataclasses import dataclass
from typing import Optional, Type

from ufo2ft.outlineCompiler import OutlineTTFCompiler
from ufo2ft.preProcessor import TTFInterpolatablePreProcessor

from .interpolatableTTFCompiler import InterpolatableTTFCompiler


@dataclass
class VariableTTFsCompiler(InterpolatableTTFCompiler):
    preProcessorClass: Type = TTFInterpolatablePreProcessor
    outlineCompilerClass: Type = OutlineTTFCompiler
    convertCubics: bool = True
    cubicConversionError: Optional[float] = None
    reverseDirection: bool = True
    flattenComponents: bool = False
    excludeVariationTables: tuple = ()
    optimizeGvar: bool = True
    colrAutoClipBoxes: bool = False
    autoUseMyMetrics: bool = True
    dropImpliedOnCurves: bool = False
    allQuadratic: bool = True
    variableFeatures: bool = True



import os
import random
import time

import pytest

from ufo2ft.fontInfoData import (
    dateStringToTimeValue,
    getAttrWithFallback,
    normalizeStringForPostscript,
)


@pytest.fixture
def info(InfoClass):
    self = InfoClass()
    self.familyName = "Family Name"
    self.styleName = "Style Name"
    self.unitsPerEm = 1000
    self.descender = -250
    self.xHeight = 450
    self.capHeight = 600
    self.ascender = 650
    self.italicAngle = 0
    return self


class GetAttrWithFallbackTest:
    @pytest.mark.parametrize(
        "infoDict,expected",
        [
            # no styleMapFamilyName, no styleMapStyleName
            (
                {},
                {
                    "familyName": "Family Name",
                    "styleName": "Style Name",
                    "styleMapFamilyName": "Family Name Style Name",
                    "styleMapStyleName": "regular",
                    "openTypeNamePreferredFamilyName": "Family Name",
                    "openTypeNamePreferredSubfamilyName": "Style Name",
                },
            ),
            # no styleMapStyleName
            (
                {"styleMapFamilyName": "Style Map Family Name"},
                {
                    "styleMapFamilyName": "Style Map Family Name",
                    "styleMapStyleName": "regular",
                    "openTypeNamePreferredFamilyName": "Family Name",
                    "openTypeNamePreferredSubfamilyName": "Style Name",
                },
            ),
            # no styleMapFamilyName, no styleMapStyleName but styleName="Regular"
            (
                {"styleName": "Regular"},
                {
                    "familyName": "Family Name",
                    "styleName": "Regular",
                    "styleMapFamilyName": "Family Name",
                    "styleMapStyleName": "regular",
                    "openTypeNamePreferredFamilyName": "Family Name",
                    "openTypeNamePreferredSubfamilyName": "Regular",
                },
            ),
            # no styleMapFamilyName but styleName="Regular"
            (
                {"styleName": "Regular", "styleMapStyleName": "regular"},
                {
                    "styleMapFamilyName": "Family Name",
                    "styleMapStyleName": "regular",
                    "openTypeNamePreferredFamilyName": "Family Name",
                    "openTypeNamePreferredSubfamilyName": "Regular",
                },
            ),
            # no styleMapStyleName but styleName="Regular"
            (
                {"styleName": "Regular", "styleMapFamilyName": "Style Map Family Name"},
                {
                    "styleMapFamilyName": "Style Map Family Name",
                    "styleMapStyleName": "regular",
                    "openTypeNamePreferredFamilyName": "Family Name",
                    "openTypeNamePreferredSubfamilyName": "Regular",
                },
            ),
            # no styleMapFamilyName, no styleMapStyleName but styleName="Bold"
            (
                {"styleName": "Bold"},
                {
                    "familyName": "Family Name",
                    "styleName": "Bold",
                    "styleMapFamilyName": "Family Name",
                    "styleMapStyleName": "bold",
                    "openTypeNamePreferredFamilyName": "Family Name",
                    "openTypeNamePreferredSubfamilyName": "Bold",
                },
            ),
        ],
    )
    def test_family_and_style_names(self, info, infoDict, expected):
        for key, value in infoDict.items():
            setattr(info, key, value)
        for key, value in expected.items():
            assert getAttrWithFallback(info, key) == value

    def test_redundant_metadata(self, info):
        assert getAttrWithFallback(info, "openTypeNameVersion") == "Version 0.000"

        info.versionMinor = 1
        info.versionMajor = 1
        assert getAttrWithFallback(info, "openTypeNameVersion") == "Version 1.001"

        assert (
            getAttrWithFallback(info, "openTypeNameUniqueID")
            == "1.001;NONE;FamilyName-StyleName"
        )

        assert getAttrWithFallback(info, "postscriptSlantAngle") == 0

    def test_unecessary_metadata(self, info):
        assert getAttrWithFallback(info, "postscriptWeightName") is None

        info.postscriptWeightName = "Normal"
        assert getAttrWithFallback(info, "postscriptWeightName") == "Normal"

    def test_vertical_metrics(self, info):
        assert getAttrWithFallback(info, "openTypeHheaAscender") == 950
        assert getAttrWithFallback(info, "openTypeHheaDescender") == -250

        assert getAttrWithFallback(info, "openTypeOS2TypoAscender") == 650
        assert getAttrWithFallback(info, "openTypeOS2TypoDescender") == -250
        assert getAttrWithFallback(info, "openTypeOS2WinAscent") == 950
        assert getAttrWithFallback(info, "openTypeOS2WinDescent") == 250

    def test_caret_slope(self, info):
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRise") == 1000
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRun") == 0

        info.italicAngle = -12
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRise") == 1000
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRun") == 213

        info.italicAngle = 12
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRise") == 1000
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRun") == -213

        info.openTypeHheaCaretSlopeRise = 2048
        assert info.openTypeHheaCaretSlopeRun is None
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRise") == 2048
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRun") == -435

        info.openTypeHheaCaretSlopeRise = None
        info.openTypeHheaCaretSlopeRun = 200
        assert info.openTypeHheaCaretSlopeRise is None
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRise") == -941
        assert getAttrWithFallback(info, "openTypeHheaCaretSlopeRun") == 200

    def test_head_created(self, info):
        os.environ["SOURCE_DATE_EPOCH"] = "1514485183"
        try:
            assert (
                getAttrWithFallback(info, "openTypeHeadCreated")
                == "2017/12/28 18:19:43"
            )
        finally:
            del os.environ["SOURCE_DATE_EPOCH"]
        assert getAttrWithFallback(info, "openTypeHeadCreated") != "2017/12/28 18:19:43"

    def test_empty_info(self, InfoClass):
        info = InfoClass()
        assert getAttrWithFallback(info, "familyName") == "New Font"
        assert getAttrWithFallback(info, "styleName") == "Regular"
        assert getAttrWithFallback(info, "unitsPerEm") == 1000
        assert getAttrWithFallback(info, "ascender") == 800
        assert getAttrWithFallback(info, "capHeight") == 700
        assert getAttrWithFallback(info, "xHeight") == 500
        assert getAttrWithFallback(info, "descender") == -200

    def test_empty_info_2048(self, InfoClass):
        info = InfoClass()
        info.unitsPerEm = 2048
        assert getAttrWithFallback(info, "unitsPerEm") == 2048
        assert getAttrWithFallback(info, "ascender") == 1638
        assert getAttrWithFallback(info, "capHeight") == 1434
        assert getAttrWithFallback(info, "xHeight") == 1024
        assert getAttrWithFallback(info, "descender") == -410

    def test_underline_position(self, info):
        assert getAttrWithFallback(info, "postscriptUnderlinePosition") == -75

        info.postscriptUnderlinePosition = -485
        assert getAttrWithFallback(info, "postscriptUnderlinePosition") == -485


class PostscriptBlueScaleFallbackTest:
    def test_without_blue_zones(self, info):
        postscriptBlueScale = getAttrWithFallback(info, "postscriptBlueScale")
        assert postscriptBlueScale == 0.039625

    def test_with_blue_zones(self, info):
        info.postscriptBlueValues = [
            -13,
            0,
            470,
            483,
            534,
            547,
            556,
            569,
            654,
            667,
            677,
            690,
            738,
            758,
        ]
        info.postscriptOtherBlues = [-255, -245]
        postscriptBlueScale = getAttrWithFallback(info, "postscriptBlueScale")
        assert postscriptBlueScale == 0.0375


class NormalizeStringForPostscriptTest:
    def test_no_change(self):
        assert (
            normalizeStringForPostscript("Sample copyright notice.")
            == "Sample copyright notice."
        )


class DateStringToTimeValueTest:
    def test_roundtrip_random_timestamp(self):
        timestamp = random.randint(0, 10**9)
        ds = time.strftime("%Y/%m/%d %H:%M:%S", time.gmtime(timestamp))
        assert dateStringToTimeValue(ds) == timestamp


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(sys.argv))



"""Old implementation of KernFeatureWriter as of ufo2ft v2.30.0 for backward compat."""

from types import SimpleNamespace

from fontTools import unicodedata

from ufo2ft.constants import INDIC_SCRIPTS, USE_SCRIPTS
from ufo2ft.featureWriters import BaseFeatureWriter, ast
from ufo2ft.util import classifyGlyphs, quantize, unicodeScriptDirection

SIDE1_PREFIX = "public.kern1."
SIDE2_PREFIX = "public.kern2."

# In HarfBuzz the 'dist' feature is automatically enabled for these shapers:
#   src/hb-ot-shape-complex-myanmar.cc
#   src/hb-ot-shape-complex-use.cc
#   src/hb-ot-shape-complex-indic.cc
#   src/hb-ot-shape-complex-khmer.cc
# We derived the list of scripts associated to each dist-enabled shaper from
# `hb_ot_shape_complex_categorize` in src/hb-ot-shape-complex-private.hh
DIST_ENABLED_SCRIPTS = set(INDIC_SCRIPTS) | set(["Khmr", "Mymr"]) | set(USE_SCRIPTS)

RTL_BIDI_TYPES = {"R", "AL"}
LTR_BIDI_TYPES = {"L", "AN", "EN"}


def unicodeBidiType(uv):
    """Return "R" for characters with RTL direction, or "L" for LTR (whether
    'strong' or 'weak'), or None for neutral direction.
    """
    char = chr(uv)
    bidiType = unicodedata.bidirectional(char)
    if bidiType in RTL_BIDI_TYPES:
        return "R"
    elif bidiType in LTR_BIDI_TYPES:
        return "L"
    else:
        return None


class KerningPair:
    __slots__ = ("side1", "side2", "value", "directions", "bidiTypes")

    def __init__(self, side1, side2, value, directions=None, bidiTypes=None):
        if isinstance(side1, str):
            self.side1 = ast.GlyphName(side1)
        elif isinstance(side1, ast.GlyphClassDefinition):
            self.side1 = ast.GlyphClassName(side1)
        else:
            raise AssertionError(side1)

        if isinstance(side2, str):
            self.side2 = ast.GlyphName(side2)
        elif isinstance(side2, ast.GlyphClassDefinition):
            self.side2 = ast.GlyphClassName(side2)
        else:
            raise AssertionError(side2)

        self.value = value
        self.directions = directions or set()
        self.bidiTypes = bidiTypes or set()

    @property
    def firstIsClass(self):
        return isinstance(self.side1, ast.GlyphClassName)

    @property
    def secondIsClass(self):
        return isinstance(self.side2, ast.GlyphClassName)

    @property
    def glyphs(self):
        if self.firstIsClass:
            classDef1 = self.side1.glyphclass
            glyphs1 = {g.asFea() for g in classDef1.glyphSet()}
        else:
            glyphs1 = {self.side1.asFea()}
        if self.secondIsClass:
            classDef2 = self.side2.glyphclass
            glyphs2 = {g.asFea() for g in classDef2.glyphSet()}
        else:
            glyphs2 = {self.side2.asFea()}
        return glyphs1 | glyphs2

    def __repr__(self):
        return "<{} {} {} {}{}{}>".format(
            self.__class__.__name__,
            self.side1,
            self.side2,
            self.value,
            " %r" % self.directions if self.directions else "",
            " %r" % self.bidiTypes if self.bidiTypes else "",
        )


class KernFeatureWriter(BaseFeatureWriter):
    """Generates a kerning feature based on groups and rules contained
    in an UFO's kerning data.

    There are currently two possible writing modes:
    2) "skip" (default) will not write anything if the features are already present;
    1) "append" will add additional lookups to an existing feature, if present,
       or it will add a new one at the end of all features.

    If the `quantization` argument is given in the filter options, the resulting
    anchors are rounded to the nearest multiple of the quantization value.
    """

    tableTag = "GPOS"
    features = frozenset(["kern", "dist"])
    options = dict(ignoreMarks=True, quantization=1)

    def setContext(self, font, feaFile, compiler=None):
        ctx = super().setContext(font, feaFile, compiler=compiler)
        ctx.gdefClasses = self.getGDEFGlyphClasses()
        ctx.kerning = self.getKerningData(font, feaFile, self.getOrderedGlyphSet())

        feaScripts = ast.getScriptLanguageSystems(feaFile)
        ctx.scriptGroups = self._groupScriptsByTagAndDirection(feaScripts)

        return ctx

    def shouldContinue(self):
        if not self.context.kerning.pairs:
            self.log.debug("No kerning data; skipped")
            return False

        if "dist" in self.context.todo and "dist" not in self.context.scriptGroups:
            self.log.debug(
                "No dist-enabled scripts defined in languagesystem "
                "statements; dist feature will not be generated"
            )
            self.context.todo.remove("dist")

        return super().shouldContinue()

    def _write(self):
        lookups = self._makeKerningLookups()
        if not lookups:
            self.log.debug("kerning lookups empty; skipped")
            return False

        features = self._makeFeatureBlocks(lookups)
        if not features:
            self.log.debug("kerning features empty; skipped")
            return False

        # extend feature file with the new generated statements
        feaFile = self.context.feaFile

        # first add the glyph class definitions
        side1Classes = self.context.kerning.side1Classes
        side2Classes = self.context.kerning.side2Classes
        newClassDefs = []
        for classes in (side1Classes, side2Classes):
            newClassDefs.extend([c for _, c in sorted(classes.items())])

        lookupGroups = []
        for _, lookupGroup in sorted(lookups.items()):
            lookupGroups.extend(lookupGroup)

        self._insert(
            feaFile=feaFile,
            classDefs=newClassDefs,
            lookups=lookupGroups,
            features=[features[tag] for tag in ["kern", "dist"] if tag in features],
        )
        return True

    @classmethod
    def getKerningData(cls, font, feaFile=None, glyphSet=None):
        side1Classes, side2Classes = cls.getKerningClasses(font, feaFile, glyphSet)
        pairs = cls.getKerningPairs(font, side1Classes, side2Classes, glyphSet)
        return SimpleNamespace(
            side1Classes=side1Classes, side2Classes=side2Classes, pairs=pairs
        )

    @staticmethod
    def getKerningGroups(font, glyphSet=None):
        if glyphSet:
            allGlyphs = set(glyphSet.keys())
        else:
            allGlyphs = set(font.keys())
        side1Groups = {}
        side2Groups = {}
        for name, members in font.groups.items():
            # prune non-existent or skipped glyphs
            members = [g for g in members if g in allGlyphs]
            if not members:
                # skip empty groups
                continue
            # skip groups without UFO3 public.kern{1,2} prefix
            if name.startswith(SIDE1_PREFIX):
                side1Groups[name] = members
            elif name.startswith(SIDE2_PREFIX):
                side2Groups[name] = members
        return side1Groups, side2Groups

    @classmethod
    def getKerningClasses(cls, font, feaFile=None, glyphSet=None):
        side1Groups, side2Groups = cls.getKerningGroups(font, glyphSet)
        side1Classes = ast.makeGlyphClassDefinitions(
            side1Groups, feaFile, stripPrefix="public."
        )
        side2Classes = ast.makeGlyphClassDefinitions(
            side2Groups, feaFile, stripPrefix="public."
        )
        return side1Classes, side2Classes

    @staticmethod
    def getKerningPairs(font, side1Classes, side2Classes, glyphSet=None):
        if glyphSet:
            allGlyphs = set(glyphSet.keys())
        else:
            allGlyphs = set(font.keys())
        kerning = font.kerning

        pairsByFlags = {}
        for side1, side2 in kerning:
            # filter out pairs that reference missing groups or glyphs
            if side1 not in side1Classes and side1 not in allGlyphs:
                continue
            if side2 not in side2Classes and side2 not in allGlyphs:
                continue
            flags = (side1 in side1Classes, side2 in side2Classes)
            pairsByFlags.setdefault(flags, set()).add((side1, side2))

        result = []
        for flags, pairs in sorted(pairsByFlags.items()):
            for side1, side2 in sorted(pairs):
                value = kerning[side1, side2]
                if all(flags) and value == 0:
                    # ignore zero-valued class kern pairs
                    continue
                firstIsClass, secondIsClass = flags
                if firstIsClass:
                    side1 = side1Classes[side1]
                if secondIsClass:
                    side2 = side2Classes[side2]
                result.append(KerningPair(side1, side2, value))
        return result

    def _intersectPairs(self, attribute, glyphSets):
        allKeys = set()
        for pair in self.context.kerning.pairs:
            for key, glyphs in glyphSets.items():
                if not pair.glyphs.isdisjoint(glyphs):
                    getattr(pair, attribute).add(key)
                    allKeys.add(key)
        return allKeys

    @staticmethod
    def _groupScriptsByTagAndDirection(feaScripts):
        # Read scripts/languages defined in feaFile's 'languagesystem'
        # statements and group them by the feature tag (kern or dist)
        # they are associated with, and the global script's horizontal
        # direction (DFLT is excluded)
        scriptGroups = {}
        for scriptCode, scriptLangSys in feaScripts.items():
            if scriptCode:
                direction = unicodedata.script_horizontal_direction(scriptCode, "LTR")
            else:
                direction = "LTR"
            if scriptCode in DIST_ENABLED_SCRIPTS:
                tag = "dist"
            else:
                tag = "kern"
            scriptGroups.setdefault(tag, {}).setdefault(direction, []).extend(
                scriptLangSys
            )
        return scriptGroups

    @staticmethod
    def _makePairPosRule(pair, rtl=False, quantization=1):
        enumerated = pair.firstIsClass ^ pair.secondIsClass
        value = quantize(pair.value, quantization)
        if rtl and "L" in pair.bidiTypes:
            # numbers are always shaped LTR even in RTL scripts
            rtl = False
        valuerecord = ast.ValueRecord(
            xPlacement=value if rtl else None,
            yPlacement=0 if rtl else None,
            xAdvance=value,
            yAdvance=0 if rtl else None,
        )
        return ast.PairPosStatement(
            glyphs1=pair.side1,
            valuerecord1=valuerecord,
            glyphs2=pair.side2,
            valuerecord2=None,
            enumerated=enumerated,
        )

    def _makeKerningLookup(
        self, name, pairs, exclude=None, rtl=False, ignoreMarks=True
    ):
        assert pairs
        rules = []
        for pair in pairs:
            if exclude is not None and exclude(pair):
                self.log.debug("pair excluded from '%s' lookup: %r", name, pair)
                continue
            rules.append(
                self._makePairPosRule(
                    pair, rtl=rtl, quantization=self.options.quantization
                )
            )

        if rules:
            lookup = ast.LookupBlock(name)
            if ignoreMarks and self.options.ignoreMarks:
                lookup.statements.append(ast.makeLookupFlag("IgnoreMarks"))
            lookup.statements.extend(rules)
            return lookup

    def _makeKerningLookups(self):
        cmap = self.makeUnicodeToGlyphNameMapping()
        if any(unicodeScriptDirection(uv) == "RTL" for uv in cmap):
            # If there are any characters from globally RTL scripts in the
            # cmap, we compile a temporary GSUB table to resolve substitutions
            # and group glyphs by script horizontal direction and bidirectional
            # type. We then mark each kerning pair with these properties when
            # any of the glyphs involved in a pair intersects these groups.
            gsub = self.compileGSUB()
            extras = self.extraSubstitutions()
            dirGlyphs = classifyGlyphs(unicodeScriptDirection, cmap, gsub, extras)
            directions = self._intersectPairs("directions", dirGlyphs)
            shouldSplit = "RTL" in directions
            if shouldSplit:
                bidiGlyphs = classifyGlyphs(unicodeBidiType, cmap, gsub, extras)
                self._intersectPairs("bidiTypes", bidiGlyphs)
        else:
            shouldSplit = False

        marks = self.context.gdefClasses.mark
        lookups = {}
        if shouldSplit:
            # make one DFLT lookup with script-agnostic characters, and two
            # LTR/RTL lookups excluding pairs from the opposite group.
            # We drop kerning pairs with ambiguous direction: i.e. those containing
            # glyphs from scripts with different overall horizontal direction, or
            # glyphs with incompatible bidirectional type (e.g. arabic letters vs
            # arabic numerals).
            pairs = []
            for pair in self.context.kerning.pairs:
                if ("RTL" in pair.directions and "LTR" in pair.directions) or (
                    "R" in pair.bidiTypes and "L" in pair.bidiTypes
                ):
                    self.log.warning(
                        "skipped kern pair with ambiguous direction: %r", pair
                    )
                    continue
                pairs.append(pair)
            if not pairs:
                return lookups

            if self.options.ignoreMarks:
                # If there are pairs with a mix of mark/base then the IgnoreMarks
                # flag is unnecessary and should not be set
                basePairs, markPairs = self._splitBaseAndMarkPairs(pairs, marks)
                if basePairs:
                    self._makeSplitDirectionKernLookups(lookups, basePairs)
                if markPairs:
                    self._makeSplitDirectionKernLookups(
                        lookups, markPairs, ignoreMarks=False, suffix="_marks"
                    )
            else:
                self._makeSplitDirectionKernLookups(lookups, pairs)
        else:
            # only make a single (implicitly LTR) lookup including all base/base pairs
            # and a single lookup including all base/mark pairs (if any)
            pairs = self.context.kerning.pairs
            if self.options.ignoreMarks:
                basePairs, markPairs = self._splitBaseAndMarkPairs(pairs, marks)
                lookups["LTR"] = []
                if basePairs:
                    lookups["LTR"].append(
                        self._makeKerningLookup("kern_ltr", basePairs)
                    )
                if markPairs:
                    lookups["LTR"].append(
                        self._makeKerningLookup(
                            "kern_ltr_marks", markPairs, ignoreMarks=False
                        )
                    )
            else:
                lookups["LTR"] = [self._makeKerningLookup("kern_ltr", pairs)]
        return lookups

    def _splitBaseAndMarkPairs(self, pairs, marks):
        basePairs, markPairs = [], []
        if marks:
            for pair in pairs:
                if any(glyph in marks for glyph in pair.glyphs):
                    markPairs.append(pair)
                else:
                    basePairs.append(pair)
        else:
            basePairs[:] = pairs
        return basePairs, markPairs

    def _makeSplitDirectionKernLookups(
        self, lookups, pairs, ignoreMarks=True, suffix=""
    ):
        dfltKern = self._makeKerningLookup(
            "kern_dflt" + suffix,
            pairs,
            exclude=(lambda pair: {"LTR", "RTL"}.intersection(pair.directions)),
            rtl=False,
            ignoreMarks=ignoreMarks,
        )
        if dfltKern:
            lookups.setdefault("DFLT", []).append(dfltKern)

        ltrKern = self._makeKerningLookup(
            "kern_ltr" + suffix,
            pairs,
            exclude=(lambda pair: not pair.directions or "RTL" in pair.directions),
            rtl=False,
            ignoreMarks=ignoreMarks,
        )
        if ltrKern:
            lookups.setdefault("LTR", []).append(ltrKern)

        rtlKern = self._makeKerningLookup(
            "kern_rtl" + suffix,
            pairs,
            exclude=(lambda pair: not pair.directions or "LTR" in pair.directions),
            rtl=True,
            ignoreMarks=ignoreMarks,
        )
        if rtlKern:
            lookups.setdefault("RTL", []).append(rtlKern)

    def _makeFeatureBlocks(self, lookups):
        features = {}
        if "kern" in self.context.todo:
            kern = ast.FeatureBlock("kern")
            self._registerKernLookups(kern, lookups)
            if kern.statements:
                features["kern"] = kern
        if "dist" in self.context.todo:
            dist = ast.FeatureBlock("dist")
            self._registerDistLookups(dist, lookups)
            if dist.statements:
                features["dist"] = dist
        return features

    def _registerKernLookups(self, feature, lookups):
        if "DFLT" in lookups:
            ast.addLookupReferences(feature, lookups["DFLT"])

        scriptGroups = self.context.scriptGroups
        if "dist" in self.context.todo:
            distScripts = scriptGroups["dist"]
        else:
            distScripts = {}
        kernScripts = scriptGroups.get("kern", {})
        ltrScripts = kernScripts.get("LTR", [])
        rtlScripts = kernScripts.get("RTL", [])

        ltrLookups = lookups.get("LTR")
        rtlLookups = lookups.get("RTL")
        if ltrLookups and rtlLookups:
            if ltrScripts and rtlScripts:
                for script, langs in ltrScripts:
                    ast.addLookupReferences(feature, ltrLookups, script, langs)
                for script, langs in rtlScripts:
                    ast.addLookupReferences(feature, rtlLookups, script, langs)
            elif ltrScripts:
                ast.addLookupReferences(feature, rtlLookups, script="DFLT")
                for script, langs in ltrScripts:
                    ast.addLookupReferences(feature, ltrLookups, script, langs)
            elif rtlScripts:
                ast.addLookupReferences(feature, ltrLookups, script="DFLT")
                for script, langs in rtlScripts:
                    ast.addLookupReferences(feature, rtlLookups, script, langs)
            else:
                if not (distScripts.get("LTR") and distScripts.get("RTL")):
                    raise ValueError(
                        "cannot use DFLT script for both LTR and RTL kern "
                        "lookups; add 'languagesystems' to features for at "
                        "least one LTR or RTL script using the kern feature"
                    )
        elif ltrLookups:
            if not (rtlScripts or distScripts):
                ast.addLookupReferences(feature, ltrLookups)
            else:
                ast.addLookupReferences(feature, ltrLookups, script="DFLT")
                for script, langs in ltrScripts:
                    ast.addLookupReferences(feature, ltrLookups, script, langs)
        elif rtlLookups:
            if not (ltrScripts or distScripts):
                ast.addLookupReferences(feature, rtlLookups)
            else:
                ast.addLookupReferences(feature, rtlLookups, script="DFLT")
                for script, langs in rtlScripts:
                    ast.addLookupReferences(feature, rtlLookups, script, langs)

    def _registerDistLookups(self, feature, lookups):
        scripts = self.context.scriptGroups["dist"]
        ltrLookups = lookups.get("LTR")
        if ltrLookups:
            for script, langs in scripts.get("LTR", []):
                ast.addLookupReferences(feature, ltrLookups, script, langs)
        rtlLookups = lookups.get("RTL")
        if rtlLookups:
            for script, langs in scripts.get("RTL", []):
                ast.addLookupReferences(feature, rtlLookups, script, langs)



from __future__ import annotations

import logging
import sys
from types import SimpleNamespace
from typing import TYPE_CHECKING, FrozenSet, Tuple

from fontTools.misc.loggingTools import Timer

from ufo2ft.util import (
    _getNewGlyphFactory,
    _GlyphSet,
    _LazyFontName,
    getMaxComponentDepth,
    zip_strict,
)

if TYPE_CHECKING:
    from typing import Any, TypeAlias

    from ufoLib2.objects import Font, Glyph

    from ufo2ft.instantiator import Instantiator, InterpolatedLayer

# reuse the "ufo2ft.filters" logger
logger = logging.getLogger("ufo2ft.filters")

# library-level logger specialized for timing info which apps like fontmake
# can selectively configure
timing_logger = logging.getLogger("ufo2ft.timer")


class BaseFilter:
    # tuple of strings listing the names of required positional arguments
    # which will be set as attributes of the filter instance
    _args = ()

    # dictionary containing the names of optional keyword arguments and
    # their default values, which will be set as instance attributes
    _kwargs = {}

    # pre-filter when True, post-filter when False, meaning before or after default
    # filters
    _pre = False

    def __init__(self, *args, **kwargs):
        self.options = options = SimpleNamespace()

        num_required = len(self._args)
        num_args = len(args)
        # process positional arguments as keyword arguments
        if num_args < num_required:
            args = (
                *args,
                *(kwargs.pop(a) for a in self._args[num_args:] if a in kwargs),
            )
            num_args = len(args)
            duplicated_args = [k for k in self._args if k in kwargs]
            if duplicated_args:
                num_duplicated = len(duplicated_args)
                raise TypeError(
                    "got {} duplicated positional argument{}: {}".format(
                        num_duplicated,
                        "s" if num_duplicated > 1 else "",
                        ", ".join(duplicated_args),
                    )
                )
        # process positional arguments
        if num_args < num_required:
            missing = [repr(a) for a in self._args[num_args:]]
            num_missing = len(missing)
            raise TypeError(
                "missing {} required positional argument{}: {}".format(
                    num_missing, "s" if num_missing > 1 else "", ", ".join(missing)
                )
            )
        elif num_args > num_required:
            extra = [repr(a) for a in args[num_required:]]
            num_extra = len(extra)
            raise TypeError(
                "got {} unsupported positional argument{}: {}".format(
                    num_extra, "s" if num_extra > 1 else "", ", ".join(extra)
                )
            )
        for key, value in zip(self._args, args):
            setattr(options, key, value)

        # process optional keyword arguments
        for key, default in self._kwargs.items():
            setattr(options, key, kwargs.pop(key, default))

        # process special pre argument
        self.pre = kwargs.pop("pre", self._pre)

        # process special include/exclude arguments
        include = kwargs.pop("include", None)
        exclude = kwargs.pop("exclude", None)
        if include is not None and exclude is not None:
            raise ValueError("'include' and 'exclude' arguments are mutually exclusive")
        if callable(include):
            # 'include' can be a function (e.g. lambda) that takes a
            # glyph object and returns True/False based on some test
            self.include = include
            self._include_repr = lambda: repr(include)
        elif include is not None:
            # or it can be a list of glyph names to be included
            included = set(include)
            self.include = lambda g: g.name in included
            self._include_repr = lambda: repr(include)
        elif exclude is not None:
            # alternatively one can provide a list of names to not include
            excluded = set(exclude)
            self.include = lambda g: g.name not in excluded
            self._exclude_repr = lambda: repr(exclude)
        else:
            # by default, all glyphs are included
            self.include = lambda g: True

        # raise if any unsupported keyword arguments
        if kwargs:
            num_left = len(kwargs)
            raise TypeError(
                "got {}unsupported keyword argument{}: {}".format(
                    "an " if num_left == 1 else "",
                    "s" if len(kwargs) > 1 else "",
                    ", ".join(f"'{k}'" for k in kwargs),
                )
            )

        # run the filter's custom initialization code
        self.start()

    def __repr__(self):
        items = []
        if self._args:
            items.append(
                ", ".join(repr(getattr(self.options, arg)) for arg in self._args)
            )
        if self._kwargs:
            items.append(
                ", ".join(
                    "{}={!r}".format(k, getattr(self.options, k))
                    for k in sorted(self._kwargs)
                )
            )
        if hasattr(self, "_include_repr"):
            items.append(f"include={self._include_repr()}")
        elif hasattr(self, "_exclude_repr"):
            items.append(f"exclude={self._exclude_repr()}")
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def start(self):
        """Subclasses can perform here custom initialization code."""
        pass

    def set_context(self, font, glyphSet):
        """Populate a `self.context` namespace, which is reset before each
        new filter call.

        Subclasses can override this to provide contextual information
        which depends on other data in the font that is not available in
        the glyphs objects currently being filtered, or set any other
        temporary attributes.

        The default implementation simply sets the current font and glyphSet,
        and initializes an empty set that keeps track of the names of the
        glyphs that were modified.

        Returns the namespace instance.
        """
        self.context = SimpleNamespace(font=font, glyphSet=glyphSet)
        self.context.modified = set()
        proto = font.layers.defaultLayer.instantiateGlyphObject()
        self.context.glyphFactory = _getNewGlyphFactory(proto)
        return self.context

    def filter(self, glyph):
        """This is where the filter is applied to a single glyph.
        Subclasses must override this method, and return True
        when the glyph was modified.
        """
        raise NotImplementedError

    @property
    def name(self):
        return self.__class__.__name__

    def __call__(self, font, glyphSet=None):
        """Run this filter on all the included glyphs.
        Return the set of glyph names that were modified, if any.

        If `glyphSet` (dict) argument is provided, run the filter on
        the glyphs contained therein (which may be copies).
        Otherwise, run the filter in-place on the font's default
        glyph set.
        """
        fontName = _LazyFontName(font)
        if glyphSet is not None and getattr(glyphSet, "name", None):
            logger.info("Running %s on %s-%s", self.name, fontName, glyphSet.name)
        else:
            logger.info("Running %s on %s", self.name, fontName)

        if glyphSet is None:
            glyphSet = _GlyphSet.from_layer(font)

        context = self.set_context(font, glyphSet)

        filter_ = self.filter
        include = self.include
        modified = context.modified

        # process composite glyphs in decreasing component depth order (i.e. composites
        # with more deeply nested components before shallower ones) to avoid
        # order-dependent interferences while filtering glyphs with nested components
        # https://github.com/googlefonts/ufo2ft/issues/621
        orderedGlyphs = sorted(
            glyphSet.keys(), key=lambda g: -getMaxComponentDepth(glyphSet[g], glyphSet)
        )

        with Timer() as t:
            for glyphName in orderedGlyphs:
                if glyphName in modified:
                    continue
                glyph = glyphSet[glyphName]
                if include(glyph) and filter_(glyph):
                    modified.add(glyphName)

        num = len(modified)
        if num > 0:
            timing_logger.debug(
                "Took %.3fs to run %s on %d glyph%s",
                t,
                self.name,
                len(modified),
                "" if num == 1 else "s",
            )
        return modified

    @classmethod
    def getInterpolatableFilterClass(cls) -> BaseIFilter | None:
        """Return interpolatable filter class if one is found in the same module.

        We search for a class with the same name and the 'IFilter' suffix
        (where the 'I' stands for "interpolatable").

        Subclasses can override this if they wish to use a different class name
        or module.
        """
        module = sys.modules[cls.__module__]
        filter_name = cls.__name__
        if filter_name.endswith("Filter"):
            filter_name = filter_name[:-6]
        ifilter_name = f"{filter_name}IFilter"
        return getattr(module, ifilter_name, None)


HashableLocation: TypeAlias = FrozenSet[Tuple[str, float]]


class BaseIFilter(BaseFilter):
    """Interpolatable variant that zips through mutliple glyphs at a time."""

    def set_context(
        self,
        fonts: list[Font],
        glyphSets: list[dict[str, Glyph]],
        instantiator: Instantiator | None = None,
        **kwargs: dict[str, Any],
    ) -> SimpleNamespace:
        """Populate a `self.context` namespace, which is reset before each
        new filter call.

        Subclasses can override this to provide contextual information
        which depends on other data in the fonts that is not available in
        the glyphs objects currently being filtered, or set any other
        temporary attributes.

        The default implementation simply sets the current fonts, glyphSets,
        and optional instantiator and initializes an empty set that keeps track
        of the names of the glyphs that were modified.

        Any extra keyword arguments are passed to the context namespace.

        Returns the namespace instance.
        """
        assert len(fonts) == len(glyphSets)
        self.context = SimpleNamespace(
            fonts=fonts,
            glyphSets=glyphSets,
            instantiator=instantiator,
            **kwargs,
        )
        self.context.modified = set()
        # this is used to memoize locationsFromComponentGlyphs method below, to avoid
        # redoing the same work over and over again (especially when font has loads of
        # masters and many nested components).
        self.context.componentLocations = {}
        proto = fonts[0].layers.defaultLayer.instantiateGlyphObject()
        self.context.glyphFactory = _getNewGlyphFactory(proto)
        return self.context

    def filter(self, glyphName: str, glyphs: list) -> bool:
        """This is where the filter is applied to a set of interpolatable glyphs.

        Subclasses must override this method, and return True
        when the glyph was modified.
        """
        raise NotImplementedError

    def __call__(
        self,
        fonts: list[Font],
        glyphSets: list[dict[str, Glyph]] | None = None,
        instantiator: Instantiator | None = None,
        **kwargs: dict[str, Any],
    ) -> set[str]:
        """Run this filter on all the included glyphs from the given glyphSets.
        Return the set of glyph names that were modified, if any.

        If `glyphSets` (list[dict]) argument is provided, run the filter on
        the glyphs contained therein (which may be copies).
        Otherwise, run the filter in-place on the fonts' default
        glyph sets.

        The `instantiator` optional argument allows interpolatable filters to
        generate glyph instances on demand at any location in the designspace.

        Any extra keyword arguments are passed on to the `set_context` method.
        """
        logger.info("Running interpolatable %s", self.name)

        if glyphSets is None:
            glyphSets = [_GlyphSet.from_layer(font) for font in fonts]

        context = self.set_context(fonts, glyphSets, instantiator, **kwargs)

        filter_ = self.filter
        include = self.include
        modified = context.modified

        # process composite glyphs in decreasing component depth order (i.e. composites
        # with more deeply nested components before shallower ones) to avoid
        # order-dependent interferences while filtering glyphs with nested components
        # https://github.com/googlefonts/ufo2ft/issues/621
        allGlyphNames = set.union(*(set(glyphSet.keys()) for glyphSet in glyphSets))

        def comp_depth(g):
            for glyphSet in glyphSets:
                if g in glyphSet:
                    return -getMaxComponentDepth(glyphSet[g], glyphSet)
            raise AssertionError

        orderedGlyphs = sorted(allGlyphNames, key=comp_depth)

        with Timer() as t:
            for glyphName in orderedGlyphs:
                if glyphName in modified:
                    continue
                glyphs = [
                    glyphSet[glyphName]
                    for glyphSet in glyphSets
                    if glyphName in glyphSet
                ]
                if any(include(g) for g in glyphs) and filter_(glyphName, glyphs):
                    modified.add(glyphName)

        num = len(modified)
        if num > 0:
            timing_logger.debug(
                "Took %.3fs to run %s on %d glyph%s",
                t,
                self.name,
                len(modified),
                "" if num == 1 else "s",
            )
        return modified

    @classmethod
    def getInterpolatableFilterClass(cls) -> "BaseIFilter" | None:
        """Return the same class as self."""
        return cls  # no-op

    def getDefaultFont(self) -> Font:
        if self.context.instantiator is not None:
            return self.context.fonts[self.context.instantiator.default_source_idx]
        else:
            # as good a guess as any...
            return self.context.fonts[0]

    def getDefaultGlyphSet(self) -> dict[str, Glyph]:
        """Return the current glyphSet corresponding to the default location."""
        if self.context.instantiator is not None:
            default_idx = self.context.instantiator.default_source_idx
            for i, glyphSet in enumerate(self.context.glyphSets):
                if i == default_idx:
                    return glyphSet
            else:
                raise AssertionError("No default source?!")
        else:
            # we don't have enough info to know which glyphSet corresponds
            # to the default source location so we just guess it's going to
            # be the larger one given it contains all glyphs by definition.
            return max(self.context.glyphSets, key=lambda glyphSet: len(glyphSet))

    def getInterpolatedLayers(self) -> list[InterpolatedLayer] | list[None]:
        """Return InterpolatedLayers at source locations or Nones if no Instantiator."""
        if self.context.instantiator is not None:
            return self.context.instantiator.interpolated_layers
        else:
            return [None] * len(self.context.glyphSets)

    @staticmethod
    def hashableLocation(location: dict[str, float]) -> HashableLocation:
        """Convert (normalized) location dict to a hashable set of tuples."""
        return frozenset((k, v) for k, v in location.items() if v != 0.0)

    def glyphSourceLocations(self, glyphName) -> set[HashableLocation]:
        """Return locations of all the sources that have a glyph."""
        assert self.context.instantiator is not None
        return {
            self.hashableLocation(location)
            for glyphSet, location in zip_strict(
                self.context.glyphSets, self.context.instantiator.source_locations
            )
            if glyphName in glyphSet
        }

    def locationsFromComponentGlyphs(
        self,
        glyphName: str,
        include: set[str] | None = None,
    ) -> set[HashableLocation]:
        """Return locations from all the components' base glyphs, recursively."""
        logger.debug("Gathering all locations from component glyphs: %s", glyphName)
        assert self.context.instantiator is not None
        locations = set()
        cache = self.context.componentLocations
        for glyphSet in self.context.glyphSets:
            if glyphName in glyphSet:
                glyph = glyphSet[glyphName]
                for component in glyph.components:
                    baseGlyph = component.baseGlyph
                    if include is None or baseGlyph in include:
                        locations |= self.glyphSourceLocations(baseGlyph)
                        # using ternary operator instead of cache.setdefault because
                        # the latter always evaluates the second argument, whereas
                        # I want it to be lazy to avoid recursing too often.
                        locations |= (
                            cache[baseGlyph]
                            if baseGlyph in cache
                            else cache.setdefault(
                                baseGlyph,
                                self.locationsFromComponentGlyphs(baseGlyph, include),
                            )
                        )
        return locations

    def ensureCompositeDefinedAtComponentLocations(
        self,
        glyphName: str,
        include: set[str] | None = None,
    ):
        """Ensure the composite glyph is defined at all its components' locations.

        The Instantiator is used to interpolate the glyph at the missing locations.
        If we have no Instantiator, we can't interpolate so this does nothing.
        """
        if self.context.instantiator is None:
            return

        haveLocations = self.glyphSourceLocations(glyphName)
        needLocations = self.locationsFromComponentGlyphs(glyphName, include)
        locationsToAdd = needLocations - haveLocations
        if locationsToAdd:
            for glyphSet, interpolatedLayer in zip_strict(
                self.context.glyphSets, self.context.instantiator.interpolated_layers
            ):
                if self.hashableLocation(interpolatedLayer.location) in locationsToAdd:
                    assert glyphName not in glyphSet
                    logger.debug(
                        "Interpolating composite glyph %r at %s",
                        glyphName,
                        interpolatedLayer.location,
                    )
                    glyphSet[glyphName] = interpolatedLayer[glyphName]



from types import SimpleNamespace

import pytest
from fontTools.misc.loggingTools import CapturingLogHandler

from ufo2ft.filters import (
    FILTERS_KEY,
    BaseFilter,
    getFilterClass,
    loadFilterFromString,
    loadFilters,
    logger,
)

from ..testSupport import _TempModule


class FooBarFilter(BaseFilter):
    """A filter that does nothing."""

    _args = ("a", "b")
    _kwargs = {"c": 0}

    def filter(self, glyph):
        return False


@pytest.fixture(scope="module", autouse=True)
def fooBar():
    """Make a temporary 'ufo2ft.filters.fooBar' module containing a
    'FooBarFilter' class for testing the filter loading machinery.
    """
    with _TempModule("ufo2ft.filters.fooBar") as temp_module:
        temp_module.module.__dict__["FooBarFilter"] = FooBarFilter
        yield


def test_getFilterClass():
    assert getFilterClass("Foo Bar") == FooBarFilter
    assert getFilterClass("FooBar") == FooBarFilter
    assert getFilterClass("fooBar") == FooBarFilter
    with pytest.raises(ImportError):
        getFilterClass("Baz")
    with pytest.raises(ImportError):
        getFilterClass("FooBarFilter")

    with _TempModule("myfilters"), _TempModule("myfilters.fooBar") as temp_module:
        with pytest.raises(AttributeError):
            # this fails because `myfilters.fooBar` module does not
            # have a `FooBarFilter` class
            getFilterClass("Foo Bar", pkg="myfilters")

        temp_module.module.__dict__["FooBarFilter"] = FooBarFilter

        # this will attempt to import the `FooBarFilter` class from the
        # `myfilters.fooBar` module
        assert getFilterClass("Foo Bar", pkg="myfilters") == FooBarFilter


class MockFont(SimpleNamespace):
    pass


class MockGlyph(SimpleNamespace):
    pass


def test_loadFilters_empty():
    ufo = MockFont(lib={})
    assert FILTERS_KEY not in ufo.lib
    assert loadFilters(ufo) == ([], [])


@pytest.fixture
def ufo():
    ufo = MockFont(lib={})
    ufo.lib[FILTERS_KEY] = [{"name": "Foo Bar", "args": ["foo", "bar"]}]
    return ufo


def test_loadFilters_pre(ufo):
    ufo.lib[FILTERS_KEY][0]["pre"] = True
    pre, post = loadFilters(ufo)
    assert len(pre) == 1
    assert not post
    assert isinstance(pre[0], FooBarFilter)


def test_loadFilters_custom_namespace(ufo):
    ufo.lib[FILTERS_KEY][0]["name"] = "Self Destruct"
    ufo.lib[FILTERS_KEY][0]["namespace"] = "my_dangerous_filters"

    class SelfDestructFilter(FooBarFilter):
        def filter(self, glyph):
            # Don't try this at home!!! LOL :)
            # shutil.rmtree(os.path.expanduser("~"))
            return True

    with _TempModule("my_dangerous_filters"), _TempModule(
        "my_dangerous_filters.selfDestruct"
    ) as temp:
        temp.module.__dict__["SelfDestructFilter"] = SelfDestructFilter

        _, [filter_obj] = loadFilters(ufo)

    assert isinstance(filter_obj, SelfDestructFilter)


def test_loadFilters_args_missing(ufo):
    del ufo.lib[FILTERS_KEY][0]["args"]

    with pytest.raises(TypeError) as exc_info:
        loadFilters(ufo)

    assert exc_info.match("missing")


def test_loadFilters_args_unsupported(ufo):
    ufo.lib[FILTERS_KEY][0]["args"].append("baz")

    with pytest.raises(TypeError) as exc_info:
        loadFilters(ufo)

    assert exc_info.match("unsupported")


def test_loadFilters_args_as_keywords(ufo):
    del ufo.lib[FILTERS_KEY][0]["args"]
    ufo.lib[FILTERS_KEY][0]["kwargs"] = {"a": "foo", "b": "bar"}

    _, [filter_obj] = loadFilters(ufo)

    assert filter_obj.options.a == "foo"
    assert filter_obj.options.b == "bar"


def test_loadFilters_args_as_duplicated_keywords(ufo):
    ufo.lib[FILTERS_KEY][0]["args"] = ["foo"]
    ufo.lib[FILTERS_KEY][0]["kwargs"] = {"a": "foo", "b": "bar"}

    with pytest.raises(TypeError) as exc_info:
        loadFilters(ufo)

    assert exc_info.match("duplicated")


def test_loadFilters_include_all(ufo):
    _, [filter_obj] = loadFilters(ufo)

    assert filter_obj.include(MockGlyph(name="hello"))
    assert filter_obj.include(MockGlyph(name="world"))


def test_loadFilters_include_list(ufo):
    ufo.lib[FILTERS_KEY][0]["include"] = ["a", "b"]

    _, [filter_obj] = loadFilters(ufo)

    assert filter_obj.include(MockGlyph(name="a"))
    assert filter_obj.include(MockGlyph(name="b"))
    assert not filter_obj.include(MockGlyph(name="c"))


def test_loadFilters_exclude_list(ufo):
    ufo.lib[FILTERS_KEY][0]["exclude"] = ["a", "b"]

    _, [filter_obj] = loadFilters(ufo)

    assert not filter_obj.include(MockGlyph(name="a"))
    assert not filter_obj.include(MockGlyph(name="b"))
    assert filter_obj.include(MockGlyph(name="c"))


def test_loadFilters_both_include_exclude(ufo):
    ufo.lib[FILTERS_KEY][0]["include"] = ["a", "b"]
    ufo.lib[FILTERS_KEY][0]["exclude"] = ["c", "d"]

    with pytest.raises(ValueError) as exc_info:
        loadFilters(ufo)

    assert exc_info.match("arguments are mutually exclusive")


def test_loadFilters_failed(ufo):
    ufo.lib[FILTERS_KEY].append(dict(name="Non Existent"))

    with CapturingLogHandler(logger, level="ERROR") as captor:
        loadFilters(ufo)

    captor.assertRegex("Failed to load filter")


def test_loadFilters_kwargs_unsupported(ufo):
    ufo.lib[FILTERS_KEY][0]["kwargs"] = {}
    ufo.lib[FILTERS_KEY][0]["kwargs"]["c"] = 1
    ufo.lib[FILTERS_KEY][0]["kwargs"]["d"] = 2  # unknown

    with pytest.raises(TypeError) as exc_info:
        loadFilters(ufo)

    assert exc_info.match("got an unsupported keyword")


VALID_SPEC_STRINGS = [
    "RemoveOverlapsFilter",
    "PropagateAnchorsFilter(include=['a', 'b', 'c'])",
    "ufo2ft.filters.fooBar::FooBarFilter(a='a', b='b', c=1)",
]


@pytest.mark.parametrize("spec", VALID_SPEC_STRINGS)
def test_loadFilterFromString(spec, ufo):
    philter = loadFilterFromString(spec)
    assert callable(philter)


def test_loadFilterFromString_args_missing(ufo):
    with pytest.raises(TypeError) as info:
        loadFilterFromString(
            "ufo2ft.filters.fooBar::FooBarFilter(a='a', c=1)",
        )
    assert info.match("missing 1 required positional argument: 'b'")

    with pytest.raises(TypeError) as info:
        loadFilterFromString(
            "ufo2ft.filters.fooBar::FooBarFilter(c=1)",
        )
    assert info.match("missing 2 required positional arguments: 'a', 'b'")


def test_BaseFilter_repr():
    class NoArgFilter(BaseFilter):
        pass

    assert repr(NoArgFilter()) == "NoArgFilter()"

    assert repr(FooBarFilter("a", "b", c=1)) == ("FooBarFilter('a', 'b', c=1)")

    assert (
        repr(FooBarFilter("c", "d", include=["x", "y"]))
        == "FooBarFilter('c', 'd', c=0, include=['x', 'y'])"
    )

    assert (
        repr(FooBarFilter("e", "f", c=2.0, exclude=("z",)))
        == "FooBarFilter('e', 'f', c=2.0, exclude=('z',))"
    )

    def f(g):
        return False

    assert repr(
        FooBarFilter("g", "h", include=f)
    ) == "FooBarFilter('g', 'h', c=0, include={})".format(repr(f))


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(sys.argv))



class Error(Exception):
    """Base exception class for all ufo2ft errors."""

    pass


class InvalidFontData(Error):
    """Raised when input font contains invalid data."""

    pass


class InvalidFeaturesData(Error):
    """Raised when input font contains invalid features data."""

    pass


class InvalidDesignSpaceData(Error):
    """Raised when input DesignSpace document contains invalid data."""

    pass



from dataclasses import dataclass
from typing import Optional, Type

from fontTools import varLib

from ufo2ft.constants import SPARSE_OTF_MASTER_TABLES, CFFOptimization
from ufo2ft.outlineCompiler import OutlineOTFCompiler
from ufo2ft.preProcessor import OTFInterpolatablePreProcessor
from ufo2ft.util import prune_unknown_kwargs

from .baseCompiler import BaseInterpolatableCompiler


@dataclass
class InterpolatableOTFCompiler(BaseInterpolatableCompiler):
    preProcessorClass: Type = OTFInterpolatablePreProcessor
    outlineCompilerClass: Type = OutlineOTFCompiler
    roundTolerance: Optional[float] = None
    optimizeCFF: CFFOptimization = CFFOptimization.NONE
    colrLayerReuse: bool = False
    colrAutoClipBoxes: bool = False
    skipFeatureCompilation: bool = False

    def compileOutlines(self, ufo, glyphSet, layerName=None):
        kwargs = prune_unknown_kwargs(self.__dict__, self.outlineCompilerClass)
        kwargs["tables"] = SPARSE_OTF_MASTER_TABLES if layerName is not None else None
        kwargs["optimizeCFF"] = CFFOptimization.NONE
        outlineCompiler = self.outlineCompilerClass(ufo, glyphSet=glyphSet, **kwargs)
        return outlineCompiler.compile()

    def _merge(self, designSpaceDoc, excludeVariationTables):
        return varLib.build_many(
            designSpaceDoc,
            exclude=excludeVariationTables,
            optimize=self.optimizeCFF >= CFFOptimization.SPECIALIZE,
            skip_vf=lambda vf_name: self.variableFontNames
            and vf_name not in self.variableFontNames,
            colr_layer_reuse=self.colrLayerReuse,
        )



import logging

import pytest
from fontTools.cu2qu.ufo import font_to_quadratic
from fontTools.pens.hashPointPen import HashPointPen
from fontTools.ttLib.tables._g_l_y_f import (
    OVERLAP_COMPOUND,
    ROUND_XY_TO_GRID,
    USE_MY_METRICS,
    flagOverlapSimple,
)
from fontTools.ttLib.ttFont import TTFont

from ufo2ft.instructionCompiler import InstructionCompiler

from .outlineCompiler_test import getpath

TRUETYPE_INSTRUCTIONS_KEY = "public.truetype.instructions"


def expect_maxp(
    font,
    maxStorage=0,
    maxFunctionDefs=0,
    maxInstructionDefs=0,
    maxStackElements=0,
    maxSizeOfInstructions=0,
    maxZones=1,
    maxTwilightPoints=0,
):
    maxp = font["maxp"]
    assert maxp.maxStorage == maxStorage
    assert maxp.maxFunctionDefs == maxFunctionDefs
    assert maxp.maxInstructionDefs == maxInstructionDefs
    assert maxp.maxStackElements == maxStackElements
    assert maxp.maxSizeOfInstructions == maxSizeOfInstructions
    assert maxp.maxZones == maxZones
    assert maxp.maxTwilightPoints == maxTwilightPoints


def get_hash_ufo(glyph, ufo):
    hash_pen = HashPointPen(glyph.width, ufo)
    glyph.drawPoints(hash_pen)
    return hash_pen.hash


def get_hash_ttf(glyph_name, ttf):
    aw, _lsb = ttf["hmtx"][glyph_name]
    gs = ttf.getGlyphSet()
    hash_pen = HashPointPen(aw, gs)
    ttf["glyf"][glyph_name].drawPoints(hash_pen, ttf["glyf"])
    return hash_pen.hash


@pytest.fixture
def quadfont():
    font = TTFont()
    font.importXML(getpath("TestFont-TTF-post3.ttx"))
    return font


@pytest.fixture
def quadufo(FontClass):
    font = FontClass(getpath("TestFont.ufo"))
    font_to_quadratic(font)
    return font


@pytest.fixture
def quaduforeversed(FontClass):
    font = FontClass(getpath("TestFont.ufo"))
    font_to_quadratic(font=font, reverse_direction=True)
    return font


@pytest.fixture
def testufo(FontClass):
    font = FontClass(getpath("TestFont.ufo"))
    del font.lib["public.postscriptNames"]
    return font


class InstructionCompilerTest:
    # _check_glyph_hash

    def test_check_glyph_hash_match(self, quaduforeversed, quadfont):
        glyph = quaduforeversed["a"]
        ufo_hash = get_hash_ufo(glyph, quaduforeversed)
        ttglyph = quadfont["glyf"]["a"]

        ic = InstructionCompiler(quaduforeversed, quadfont)
        result = ic._check_glyph_hash(glyph.name, ttglyph, ufo_hash)
        assert result

    def test_check_glyph_hash_missing(self, quaduforeversed, quadfont):
        glyph = quaduforeversed["a"]

        ic = InstructionCompiler(quaduforeversed, quadfont)
        result = ic._check_glyph_hash(
            glyph.name,
            quadfont["glyf"]["a"],
            None,
        )
        assert not result

    def test_check_glyph_hash_mismatch(self, testufo, quadfont):
        glyph = testufo["a"]
        ufo_hash = get_hash_ufo(glyph, testufo)
        ttglyph = quadfont["glyf"]["a"]

        # The contour direction is reversed in testufo vs. quadfont, so the
        # hash should not match

        ic = InstructionCompiler(testufo, quadfont)
        result = ic._check_glyph_hash(
            glyph.name,
            ttglyph,
            ufo_hash,
        )
        assert not result

    def test_check_glyph_hash_mismatch_composite(self, testufo, quadfont):
        glyph = testufo["h"]
        ufo_hash = get_hash_ufo(glyph, testufo)
        ttglyph = quadfont["glyf"]["h"]

        # The contour direction is reversed in testufo vs. quadfont, so the
        # hash should not match

        ic = InstructionCompiler(testufo, quadfont)
        result = ic._check_glyph_hash(
            glyph.name,
            ttglyph,
            ufo_hash,
        )
        assert not result

    def test_check_glyph_hash_mismatch_width(self, quaduforeversed, quadfont):
        glyph = quaduforeversed["a"]

        # Modify the glyph width in the UFO to trigger the mismatch
        glyph.width += 10

        ufo_hash = get_hash_ufo(glyph, quaduforeversed)
        ttglyph = quadfont["glyf"]["a"]

        ic = InstructionCompiler(quaduforeversed, quadfont)
        result = ic._check_glyph_hash(
            glyph.name,
            ttglyph,
            ufo_hash,
        )
        assert not result

    # _check_tt_data_format

    def test_check_tt_data_format_match_str(self):
        result = InstructionCompiler._check_tt_data_format(
            ttdata={"formatVersion": "1"},
            name="",
        )
        assert result is None

    def test_check_tt_data_format_type_error(self):
        with pytest.raises(
            TypeError,
            match=(
                "Illegal type 'int' instead of 'str' for formatVersion "
                "for instructions in location."
            ),
        ):
            InstructionCompiler._check_tt_data_format(
                ttdata={"formatVersion": 1},  # Spec requires a str
                name="location",
            )

    def test_check_tt_data_format_mismatch_str(self):
        with pytest.raises(
            NotImplementedError,
            match="Unknown formatVersion 1.5 for instructions in location.",
        ):
            InstructionCompiler._check_tt_data_format(
                ttdata={"formatVersion": "1.5"},  # Maps to the correct int
                name="location",
            )

    # _compile_program

    def test_compile_program_no_ttdata(self, quadufo):
        # UFO contains no "public.truetype.instructions" lib key
        ic = InstructionCompiler(quadufo, TTFont())
        for key, tag in (
            ("controlValueProgram", "prep"),
            ("fontProgram", "fpgm"),
        ):
            ic._compile_program(key=key, table_tag=tag)
        assert "fpgm" not in ic.otf
        assert "prep" not in ic.otf

    def test_compile_program_no_programs(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are not there. (They are optional)
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
        }
        for key, tag in (
            ("controlValueProgram", "prep"),
            ("fontProgram", "fpgm"),
        ):
            ic._compile_program(key=key, table_tag=tag)
        assert "fpgm" not in ic.otf
        assert "prep" not in ic.otf

    def test_compile_program_none(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are None.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValueProgram": None,
            "fontProgram": None,
        }
        for key, tag in (
            ("controlValueProgram", "prep"),
            ("fontProgram", "fpgm"),
        ):
            ic._compile_program(key=key, table_tag=tag)
        assert "fpgm" not in ic.otf
        assert "prep" not in ic.otf

    def test_compile_program_empty(self, quadufo, caplog):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are empty.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValueProgram": "",
            "fontProgram": "",
        }
        with caplog.at_level(logging.DEBUG, logger="ufo2ft.instructionCompiler"):
            for key, tag in (
                ("controlValueProgram", "prep"),
                ("fontProgram", "fpgm"),
            ):
                ic._compile_program(key=key, table_tag=tag)
        assert (
            "Assembly for table 'fpgm' is empty, table not added to font."
            in caplog.text
        )
        assert (
            "Assembly for table 'prep' is empty, table not added to font."
            in caplog.text
        )
        assert "fpgm" not in ic.otf
        assert "prep" not in ic.otf

    def test_compile_program(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, and the font and
        # control value programs are present.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValueProgram": "PUSHW[]\n511\nSCANCTRL[]",
            "fontProgram": "PUSHB[]\n0\nFDEF[]\nPOP[]\nENDF[]",
        }
        for key, tag in (
            ("controlValueProgram", "prep"),
            ("fontProgram", "fpgm"),
        ):
            ic._compile_program(key=key, table_tag=tag)

        assert "fpgm" in ic.otf
        assert "prep" in ic.otf

        # Check if the bytecode is correct, though this may be out of scope
        assert ic.otf["fpgm"].program.getBytecode() == b"\xb0\x00\x2C\x21\x2D"
        assert ic.otf["prep"].program.getBytecode() == b"\xb8\x01\xff\x85"

    # compileGlyphInstructions

    def test_compileGlyphInstructions_missing_glyph(self, caplog):
        # The method logs an info when trying to compile a glyph which is
        # missing in the UFO
        ic = InstructionCompiler(dict(), None)
        with caplog.at_level(logging.INFO, logger="ufo2ft.instructionCompiler"):
            ic.compileGlyphInstructions(None, "A")
        assert "Skipping compilation of instructions for glyph 'A'" in caplog.text
        # ... except for '.notdef' which is frequently generated
        with caplog.at_level(logging.INFO, logger="ufo2ft.instructionCompiler"):
            ic.compileGlyphInstructions(None, ".notdef")
        assert (
            "Skipping compilation of instructions for glyph '.notdef'"
            not in caplog.text
        )

    # _compile_tt_glyph_program

    def test_compile_tt_glyph_program_empty(self, quaduforeversed, quadfont):
        # UFO glyph contains no "public.truetype.instructions" lib key
        with pytest.raises(
            TypeError,
            match=(
                "Illegal type 'NoneType' instead of 'str' for formatVersion "
                "for instructions in glyph 'a'."
            ),
        ):
            InstructionCompiler(quaduforeversed, quadfont)._compile_tt_glyph_program(
                glyph=quaduforeversed["a"],
                ttglyph=quadfont["glyf"]["a"],
                ttdata={},
            )

    def test_compile_tt_glyph_program_no_asm(self, quaduforeversed, quadfont, caplog):
        # UFO glyph contains "public.truetype.instructions" lib key, but no
        # assembly code entry
        ic = InstructionCompiler(quaduforeversed, quadfont)

        assert not ic.otf["glyf"]["a"].isComposite()

        glyph = ic.ufo["a"]
        glyph_hash = get_hash_ufo(glyph, ic.ufo)

        with caplog.at_level(logging.ERROR, logger="ufo2ft.instructionCompiler"):
            ic._compile_tt_glyph_program(
                glyph=ic.ufo["a"],
                ttglyph=ic.otf["glyf"]["a"],
                ttdata={
                    "formatVersion": "1",
                    "id": glyph_hash,
                    # "assembly": "",
                },
            )
        assert (
            "Glyph assembly missing, glyph 'a' will have no instructions in font."
            in caplog.text
        )

    def test_compile_tt_glyph_program_empty_asm(
        self, quaduforeversed, quadfont, caplog
    ):
        # UFO glyph contains "public.truetype.instructions" lib key, but the
        # assembly code entry is empty
        ic = InstructionCompiler(quaduforeversed, quadfont)

        assert not ic.otf["glyf"]["a"].isComposite()

        glyph = ic.ufo["a"]
        glyph_hash = get_hash_ufo(glyph, ic.ufo)

        with caplog.at_level(logging.DEBUG, logger="ufo2ft.instructionCompiler"):
            ic._compile_tt_glyph_program(
                glyph=ic.ufo["a"],
                ttglyph=ic.otf["glyf"]["a"],
                ttdata={
                    "formatVersion": "1",
                    "id": glyph_hash,
                    "assembly": "",
                },
            )
        assert "Glyph 'a' has no instructions." in caplog.text
        assert not hasattr(ic.otf["glyf"]["h"], "program")

    def test_compile_tt_glyph_program_empty_asm_composite(
        self, quaduforeversed, quadfont
    ):
        # UFO glyph contains "public.truetype.instructions" lib key, but the
        # assembly code entry is empty. The glyph is a composite.
        ic = InstructionCompiler(quaduforeversed, quadfont)

        glyph = ic.ufo["h"]
        glyph_hash = get_hash_ufo(glyph, ic.ufo)

        assert ic.otf["glyf"]["h"].isComposite()

        ic._compile_tt_glyph_program(
            glyph=ic.ufo["h"],
            ttglyph=ic.otf["glyf"]["h"],
            ttdata={
                "formatVersion": "1",
                "id": glyph_hash,
                "assembly": "",
            },
        )
        # Components must not have an empty program
        assert not hasattr(ic.otf["glyf"]["h"], "program")

    def test_compile_tt_glyph_program(self, quaduforeversed, quadfont):
        # UFO glyph contains "public.truetype.instructions" lib key, and the
        # assembly code entry is present.
        ic = InstructionCompiler(quaduforeversed, quadfont)

        assert not ic.otf["glyf"]["a"].isComposite()

        glyph = ic.ufo["a"]
        glyph_hash = get_hash_ufo(glyph, ic.ufo)

        ic._compile_tt_glyph_program(
            glyph=ic.ufo["a"],
            ttglyph=ic.otf["glyf"]["a"],
            ttdata={
                "formatVersion": "1",
                "id": glyph_hash,
                "assembly": "PUSHB[]\n0\nMDAP[1]",
            },
        )
        assert ic.otf["glyf"]["a"].program.getBytecode() == b"\xb0\x00\x2f"

    def test_compile_tt_glyph_program_composite(self, quaduforeversed, quadfont):
        # UFO glyph contains "public.truetype.instructions" lib key, and the
        # assembly code entry is present. The glyph is a composite.
        name = "k"  # Name of the composite glyph
        ic = InstructionCompiler(quaduforeversed, quadfont)

        assert ic.otf["glyf"][name].isComposite()

        glyph_hash = get_hash_ufo(ic.ufo[name], ic.ufo)

        ic._compile_tt_glyph_program(
            glyph=ic.ufo[name],
            ttglyph=ic.otf["glyf"][name],
            ttdata={
                "formatVersion": "1",
                "id": glyph_hash,
                "assembly": "PUSHB[]\n0\nMDAP[1]",
            },
        )
        ttglyph = ic.otf["glyf"][name]
        assert hasattr(ttglyph, "program")
        assert ttglyph.program.getBytecode() == b"\xb0\x00\x2f"

    # _set_composite_flags

    def test_set_composite_flags_no_ttdata(self, quadufo, quadfont):
        name = "h"  # Name of the composite glyph
        ic = InstructionCompiler(quadufo, quadfont)

        glyph = quadufo[name]
        ttglyph = quadfont["glyf"][name]

        ic._set_composite_flags(
            glyph=glyph,
            ttglyph=ttglyph,
        )

        # Flags have been set by heuristics
        assert not ttglyph.components[0].flags & OVERLAP_COMPOUND
        assert ttglyph.components[0].flags & ROUND_XY_TO_GRID
        assert not ttglyph.components[0].flags & USE_MY_METRICS
        assert not ttglyph.components[1].flags & OVERLAP_COMPOUND
        assert ttglyph.components[1].flags & ROUND_XY_TO_GRID
        assert ttglyph.components[1].flags & USE_MY_METRICS

    def test_set_composite_flags_compound(self, quadufo, quadfont):
        name = "k"  # Name of the composite glyph
        ic = InstructionCompiler(quadufo, quadfont)

        glyph = quadufo[name]
        glyph.components[0].identifier = "component0"
        glyph.components[1].identifier = "component1"
        glyph.lib = {"public.truetype.overlap": True}
        ttglyph = quadfont["glyf"][name]

        ic._set_composite_flags(
            glyph=glyph,
            ttglyph=ttglyph,
        )
        # The OVERLAP_COMPOUND flag is only set on 1st component
        assert ttglyph.components[0].flags & OVERLAP_COMPOUND
        assert not ttglyph.components[1].flags & OVERLAP_COMPOUND

    def test_set_composite_flags_no_compound(self, quadufo, quadfont):
        name = "k"  # Name of the composite glyph
        ic = InstructionCompiler(quadufo, quadfont)

        glyph = quadufo[name]
        glyph.components[0].identifier = "component0"
        glyph.components[1].identifier = "component1"
        glyph.lib = {"public.truetype.overlap": False}
        ttglyph = quadfont["glyf"][name]

        ic._set_composite_flags(
            glyph=glyph,
            ttglyph=ttglyph,
        )
        assert not ttglyph.components[0].flags & OVERLAP_COMPOUND
        assert not ttglyph.components[1].flags & OVERLAP_COMPOUND

    def test_set_composite_flags(self, quadufo, quadfont):
        name = "h"  # Name of the composite glyph
        ic = InstructionCompiler(quadufo, quadfont)

        glyph = quadufo[name]
        glyph.components[0].identifier = "component0"
        glyph.components[1].identifier = "component1"
        glyph.lib = {
            "public.objectLibs": {
                "component0": {
                    "public.truetype.roundOffsetToGrid": False,
                    "public.truetype.useMyMetrics": False,
                },
                "component1": {
                    "public.truetype.roundOffsetToGrid": True,
                    "public.truetype.useMyMetrics": True,
                },
            },
        }
        ttglyph = quadfont["glyf"][name]

        ic._set_composite_flags(
            glyph=glyph,
            ttglyph=ttglyph,
        )

        assert not ttglyph.components[0].flags & OVERLAP_COMPOUND
        assert not ttglyph.components[0].flags & ROUND_XY_TO_GRID
        assert not ttglyph.components[0].flags & USE_MY_METRICS

        assert not ttglyph.components[1].flags & OVERLAP_COMPOUND
        assert ttglyph.components[1].flags & ROUND_XY_TO_GRID
        assert ttglyph.components[1].flags & USE_MY_METRICS

    def test_set_composite_flags_metrics_first_only(self, quadufo, quadfont):
        name = "h"  # Name of the composite glyph
        ic = InstructionCompiler(quadufo, quadfont)

        glyph = quadufo[name]
        glyph.components[0].identifier = "component0"
        glyph.components[1].identifier = "component1"
        glyph.lib = {
            "public.objectLibs": {
                "component0": {
                    "public.truetype.useMyMetrics": True,
                },
                "component1": {
                    "public.truetype.useMyMetrics": True,
                },
            },
        }
        ttglyph = quadfont["glyf"][name]

        ic._set_composite_flags(
            glyph=glyph,
            ttglyph=ttglyph,
        )

        # Flag on component 1 should have been ignored
        assert ttglyph.components[0].flags & USE_MY_METRICS
        assert not ttglyph.components[1].flags & USE_MY_METRICS

    def test_set_composite_flags_metrics_no_id(self, quadufo, quadfont):
        name = "h"  # Name of the composite glyph
        ic = InstructionCompiler(quadufo, quadfont)

        glyph = quadufo[name]
        # First component has no identifier
        glyph.components[0].identifier = None
        glyph.components[1].identifier = "component1"
        glyph.lib = {
            "public.objectLibs": {
                "component1": {
                    "public.truetype.useMyMetrics": False,
                },
            },
        }
        ttglyph = quadfont["glyf"][name]

        ic._set_composite_flags(
            glyph=glyph,
            ttglyph=ttglyph,
        )

        # Flag on both components should have been unset
        assert not ttglyph.components[0].flags & USE_MY_METRICS
        assert not ttglyph.components[1].flags & USE_MY_METRICS

    @pytest.mark.parametrize("autoUseMyMetrics", [True, False])
    def test_set_composite_flags_auto_use_my_metrics_warn_if_components_mismatch(
        self, quadufo, quadfont, autoUseMyMetrics, caplog
    ):
        ic = InstructionCompiler(quadufo, quadfont, autoUseMyMetrics=autoUseMyMetrics)
        name = "h"

        glyph = quadufo[name]
        ttglyph = quadfont["glyf"][name]
        assert len(glyph.components) == len(ttglyph.components)
        glyph.clearComponents()  # to produce an artificial len(components) mismatch
        assert len(glyph.components) != len(ttglyph.components)

        with caplog.at_level(logging.DEBUG, logger="ufo2ft.instructionCompiler"):
            ic._set_composite_flags(glyph=glyph, ttglyph=ttglyph)

        assert "Number of components differ" in caplog.text

    @pytest.mark.parametrize("overlap", [None, False, True])
    def test_set_simple_flags(self, quadufo, quadfont, overlap):
        ic = InstructionCompiler(quadufo, quadfont)
        name = "a"

        glyph = quadufo[name]
        if overlap is not None:
            glyph.lib = {"public.truetype.overlap": overlap}
        ttglyph = quadfont["glyf"][name]

        ic._set_simple_flags(glyph=glyph, ttglyph=ttglyph)

        if overlap:
            assert ttglyph.flags[0] & flagOverlapSimple
        else:
            assert not ttglyph.flags[0] & flagOverlapSimple

    # update_maxp

    def test_update_maxp_no_ttdata(self, quaduforeversed, quadfont):
        ic = InstructionCompiler(quaduforeversed, quadfont)

        ic.update_maxp()
        expect_maxp(ic.otf)

    def test_update_maxp(self, quaduforeversed, quadfont):
        ic = InstructionCompiler(quaduforeversed, quadfont)
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "maxStorage": 1,
            "maxFunctionDefs": 1,
            "maxInstructionDefs": 1,
            "maxStackElements": 1,
            "maxSizeOfInstructions": 1,
            "maxZones": 2,
            "maxTwilightPoints": 1,
        }
        # Make a glyph program of size 3 in "a"
        self.test_compile_tt_glyph_program(quaduforeversed, quadfont)
        ic.update_maxp()
        # maxSizeOfInstructions should be 3 because it is calculated from the font
        expect_maxp(ic.otf, 1, 1, 1, 1, 3, 2, 1)

    # setupTable_cvt

    def test_setupTable_cvt(self, quaduforeversed, quadfont):
        ic = InstructionCompiler(quaduforeversed, quadfont)
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValue": {
                "1": 500,
                "2": 750,
                "3": -250,
            },
        }
        ic.setupTable_cvt()
        assert "cvt " in ic.otf
        assert list(ic.otf["cvt "].values) == [0, 500, 750, -250]

    def test_setupTable_cvt_empty(self, quaduforeversed, quadfont):
        ic = InstructionCompiler(quaduforeversed, quadfont)
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValue": {},
        }
        ic.setupTable_cvt()
        assert "cvt " not in ic.otf

    def test_setupTable_cvt_none(self, quaduforeversed, quadfont):
        ic = InstructionCompiler(quaduforeversed, quadfont)
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValue": None,
        }
        ic.setupTable_cvt()
        assert "cvt " not in ic.otf

    def test_setupTable_cvt_missing(self, quaduforeversed, quadfont):
        ic = InstructionCompiler(quaduforeversed, quadfont)
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
        }
        ic.setupTable_cvt()
        assert "cvt " not in ic.otf

    def test_setupTable_cvt_no_ttdata(self, quaduforeversed, quadfont):
        ic = InstructionCompiler(quaduforeversed, quadfont)
        ic.setupTable_cvt()
        assert "cvt " not in ic.otf

    # setupTable_fpgm

    def test_setupTable_fpgm_no_ttdata(self, quadufo):
        # UFO contains no "public.truetype.instructions" lib key
        ic = InstructionCompiler(quadufo, TTFont())
        ic.setupTable_fpgm()
        assert "fpgm" not in ic.otf

    def test_setupTable_fpgm_no_program(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are not there. (They are optional)
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
        }
        ic.setupTable_fpgm()
        assert "fpgm" not in ic.otf

    def test_setupTable_fpgm_none(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are None.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "fontProgram": None,
        }
        ic.setupTable_fpgm()
        assert "fpgm" not in ic.otf

    def test_setupTable_fpgm_empty(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are empty.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "fontProgram": "",
        }
        ic.setupTable_fpgm()
        assert "fpgm" not in ic.otf

    def test_setupTable_fpgm(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, and the font and
        # control value programs are present.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "fontProgram": "PUSHB[]\n0\nFDEF[]\nPOP[]\nENDF[]",
        }
        ic.setupTable_fpgm()

        assert "fpgm" in ic.otf

        # Check if the bytecode is correct, though this may be out of scope
        assert ic.otf["fpgm"].program.getBytecode() == b"\xb0\x00\x2C\x21\x2D"

    # setupTable_gasp

    def test_setupTable_gasp(self, testufo):
        ic = InstructionCompiler(testufo, TTFont())
        ic.setupTable_gasp()
        assert "gasp" in ic.otf
        assert ic.otf["gasp"].gaspRange == {7: 10, 65535: 15}

    def test_compile_without_gasp(self, testufo):
        testufo.info.openTypeGaspRangeRecords = None
        ic = InstructionCompiler(testufo, TTFont())
        ic.setupTable_gasp()
        assert "gasp" not in ic.otf

    def test_compile_empty_gasp(self, testufo):
        # ignore empty gasp
        testufo.info.openTypeGaspRangeRecords = []
        ic = InstructionCompiler(testufo, TTFont())
        ic.setupTable_gasp()
        assert "gasp" not in ic.otf

    # setupTable_prep

    def test_setupTable_prep_no_ttdata(self, quadufo):
        # UFO contains no "public.truetype.instructions" lib key
        ic = InstructionCompiler(quadufo, TTFont())
        ic.setupTable_prep()
        assert "prep" not in ic.otf

    def test_setupTable_prep_no_program(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are not there. (They are optional)
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
        }
        ic.setupTable_prep()
        assert "prep" not in ic.otf

    def test_setupTable_prep_none(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are None.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValueProgram": None,
        }
        ic.setupTable_prep()
        assert "prep" not in ic.otf

    def test_setupTable_prep_empty(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, but the font and
        # control value programs are empty.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValueProgram": "",
        }
        ic.setupTable_prep()
        assert "prep" not in ic.otf

    def test_setupTable_prep(self, quadufo):
        # UFO contains the "public.truetype.instructions" lib key, and the font and
        # control value programs are present.
        ic = InstructionCompiler(quadufo, TTFont())
        ic.ufo.lib[TRUETYPE_INSTRUCTIONS_KEY] = {
            "formatVersion": "1",
            "controlValueProgram": "PUSHW[]\n511\nSCANCTRL[]",
        }
        ic.setupTable_prep()

        assert "prep" in ic.otf

        # Check if the bytecode is correct, though this may be out of scope
        assert ic.otf["prep"].program.getBytecode() == b"\xb8\x01\xff\x85"



from __future__ import annotations

from typing import TYPE_CHECKING

from ufo2ft.filters import BaseFilter, BaseIFilter
from ufo2ft.util import decomposeCompositeGlyph, zip_strict

if TYPE_CHECKING:
    from ufoLib2.objects import Glyph


class DecomposeComponentsFilter(BaseFilter):
    # pre=True so by default this is run before the RemoveOverlaps filter,
    # in case a component overlaps other contours or components, to ensure
    # the decomposed contours will be merged correctly:
    # https://github.com/googlefonts/gftools/pull/425
    _pre = True

    def filter(self, glyph: Glyph) -> bool:
        if not glyph.components:
            return False
        decomposeCompositeGlyph(glyph, self.context.glyphSet)
        return True


class DecomposeComponentsIFilter(BaseIFilter):
    _pre = True

    def filter(self, glyphName: str, glyphs: list[Glyph]) -> bool:
        if not any(glyph.components for glyph in glyphs):
            return False

        self.ensureCompositeDefinedAtComponentLocations(glyphName)

        for glyphSet, interpolatedLayer in zip_strict(
            self.context.glyphSets, self.getInterpolatedLayers()
        ):
            glyph = glyphSet.get(glyphName)
            if glyph is not None:
                decomposeCompositeGlyph(glyph, interpolatedLayer or glyphSet)
        return True
